'
' Interfacing to a Sure 2416 24*16 Dot Matrix Display Board
' With the Amicus18 BASIC compiler
'
' Version 1.1 (20-03-10)
' Corrected HT_Pixel routine
'
Reminders = Off
Warnings = Off
    Optimiser_Level = 3                 ' Maximum optimisation

$ifndef HT_CS     
    $define HT_CS PORTB.2               ' Connects to the Sure2416 CS1 pin (Active Low)
$endif
$ifndef HT_WR
    $define HT_WR PORTB.1               ' Connects to the Sure2416 WR pin  (Active Low)
$endif
$ifndef HT_RD
    $define HT_RD PORTB.4               ' Connects to the Sure2416 RD pin  (Active Low)
$endif
$ifndef HT_Data
    $define HT_Data PORTB.3             ' Connects to the Sure2416 DATA pin
$endif
'
' Constants
'
' Commands written to the chip consist of a 3 bit "ID", followed by
' either 9 bits of "Command code" or 7 bits of address, plus 4 bits of data.
'
    Symbol cHT_ID_CMD 4                 ' ID = 100 - Commands 
    Symbol cHT_ID_RD  6                 ' ID = 110 - Read RAM 
    Symbol cHT_ID_WR  5                 ' ID = 101 - Write RAM 

    Symbol cHT_CMD_SYSDIS = $00         ' CMD = 0000-0000-x     Turn off oscillator 
    Symbol cHT_CMD_SYSON  = $01         ' CMD = 0000-0001-x     Enable system oscillator 
    Symbol cHT_CMD_LEDOFF = $02         ' CMD = 0000-0010-x     LED duty cycle gen off 
    Symbol cHT_CMD_LEDON  = $03         ' CMD = 0000-0011-x     LEDs On 
    Symbol cHT_CMD_BLOFF  = $08         ' CMD = 0000-1000-x     Blink On 
    Symbol cHT_CMD_BLON   = $09         ' CMD = 0000-1001-x     Blink Off 
    Symbol cHT_CMD_SLVMD  = $10         ' CMD = 0001-0 = $x-x   Slave Mode 
    Symbol cHT_CMD_MSTMD  = $14         ' CMD = 0001-01xx-x     Master Mode 
    Symbol cHT_CMD_RCCLK  = $18         ' CMD = 0001-1 = $x-x   Use on-chip clock 
    Symbol cHT_CMD_EXTCLK = $1C         ' CMD = 0001-11xx-x     Use external clock 
    Symbol cHT_CMD_COMS00 = $20         ' CMD = 0010-ABxx-x     commons options 
    Symbol cHT_CMD_COMS01 = $24         ' CMD = 0010-ABxx-x     commons options 
    Symbol cHT_CMD_COMS10 = $28         ' CMD = 0010-ABxx-x     commons options 
    Symbol cHT_CMD_COMS11 = $2C         ' CMD = 0010-ABxx-x     commons options 
    Symbol cHT_CMD_PWM    = $A0         ' CMD = 101x-PPPP-x     PWM duty cycle 
    
    Symbol cHT_XMax = 23
    Symbol cHT_YMax = 15
    
    Symbol tPixelClear = 0                  ' Set the pixel colour to off
    Symbol tPixelSet = 1                    ' Set the pixel colour to on
    Symbol tPixelXor = 2                    ' Set the pixel colour to xor
'
' Compiler system variables. Re-used by our routines
'    
    Dim GEN As Byte System
    Dim GENH As Byte System
    Dim GEN2 As Byte System
    Dim GEN2H As Byte System
    Dim PP0 As Byte System
    Dim PP0H As Byte System
    Dim PP1 As Byte System
    Dim PP1H As Byte System
    Dim PP2 As Byte System
    Dim PP2H As Byte System
    Dim PP3 As Byte System 
    Dim PP3H As Byte System 
    Dim wHT_CharOffset As PRODL.Word               ' Alias the SFR pair PRODL and PRODH as a word variable   
    Dim wHT_FSR1 As FSR1L.Word                     ' Alias the SFR pair FSR1L and FSR1H as a word variable
'
' Routine Variables
'         
    Dim bHT_SureOutValue As Byte System
    Dim bHT_SureAddress As Byte System
    Dim bHT_WrBits As PRODL
    Dim bHT_WrFbits As PRODH
    Dim tPixelColor As Byte System                 ' Used to hold the colour of the pixel for all routines except DisplayChar  
    Dim bHT_SetPixelX As Byte System               ' Used by the Plot/Pixel subroutines to hold the X position of the pixel
    Dim bHT_SetPixelY As Byte System               ' Used by the Plot/Pixel subroutines to hold the Y position of the pixel
    Dim bHT_Address As Byte System                 ' Used by the Plot/Pixel subroutines to calculate the pixel address
    Dim bHT_BitVal As Byte System                  ' Used by the Plot/Pixel subroutines to calculate the pixel position withn memory
'    
' Declare variables used by DisplayChar and DisplayString
'    
    Dim bHT_CharX As GEN
    Dim bHT_CharY As GENH
    Dim bHT_Col As PP0
    Dim bHT_Row As PP0H                                
'    
' Declare variables used by HT_Line
'
    Dim bHT_DeltaX As GEN
    Dim bHT_DeltaY As GENH
    Dim bHT_StepX As PP0
    Dim bHT_StepY As PP0H
    Dim bHT_Deviation As PP3
    Dim bHT_Ypos_S As PP1                        
    Dim bHT_Xpos_S As PP1H                        
    Dim bHT_Ypos_E As PP2                        
    Dim bHT_Xpos_E As PP2H    
'    
' Declare variables (aliases) used by HT_Circle
'    
    Dim bHT_XX As GEN
    Dim bHT_YY As GENH
    Dim bHT_TR As PP0
    Dim bHT_DD As PP0H
    Dim bHT_QuadA As PP2
    Dim bHT_QuadB As PP2H
    Dim bHT_QuadC As PP3
    Dim bHT_QuadD As GEN2
    Dim bHT_Radius As GEN2H
'
' Keep a copy of the display controller's RAM so that we can know which bits are set without having to read the device.
'
    Dim bHT_ShadowRam[96] As Byte            ' Copy of the display's RAM
    Dim tHT_PixelColSave As Byte             ' Holds a copy of tPixelColor when impementing DispayChar
    
'-------------------------------------------------------------------------------------    
    GoTo HT_StartUp                          ' Jump over the subroutines

'-------------------------------------------------------------------------------------
' Shift out data to the Sure 2416 board
'
' Input     : bHT_WrBits holds the value to send
'           : bHT_WrFbits (WREG) holds the starting bit of the value
' Output    : None
' Notes     : None
'
$define HT_ShiftOut(pData, pFirstBit) bHT_WrBits = pData : WREG = pFirstBit : _HT_ShiftOut
_HT_ShiftOut Macro
    GoSub __HT_ShiftOut
Endm
#ifdef _HT_ShiftOut#Req
__HT_ShiftOut:
    bHT_WrFbits = WREG
    While bHT_WrFbits <> 0
        Clear HT_WR
        WREG = bHT_WrBits & bHT_WrFbits
        If WREG <> 0 Then
            Set HT_Data
        Else 
            Clear HT_Data
        EndIf
        Set HT_WR
        bHT_WrFbits = bHT_WrFbits >> 1
    Wend
    Return
#endif 
'-------------------------------------------------------------------------------------
' Send a command to the HT
'
' Input     : bHT_SureOutValue holds the command to send
' Output    : None
' Notes     : None
'
$define HT_SendCommand(pCommand) WREG = pCommand : _HT_SendCommand
_HT_SendCommand Macro
    GoSub __HT_SendCommand
Endm
#ifdef _HT_SendCommand#Req
__HT_SendCommand:
    bHT_SureOutValue = WREG
    Clear HT_CS                                 ' Select chip
    HT_ShiftOut(cHT_ID_CMD, 1 << 2)             ' Send 3 bits of ID: Commmand
    HT_ShiftOut(bHT_SureOutValue, 1 << 7)       ' Send the actual command
    HT_ShiftOut(0, 1)                           ' One extra don't care bit in commands 
    Set HT_CS                                   ' De-Select Chip
    Return
#endif
'-------------------------------------------------------------------------------------
' Send a data byte to the HT
'
' Input     : bHT_SureOutValue holds the data to send
'             bHT_SureAddress holds the address
' Output    : None
' Notes     : None
'
$define HT_SendData(pAddress,pData) bHT_SureAddress = pAddress : WREG = pData : _HT_SendData
_HT_SendData Macro
    GoSub __HT_SendData
Endm
#ifdef _HT_SendData#Req
__HT_SendData:
    bHT_SureOutValue = WREG                     ' Transfer WREG into it's working variable
    Clear HT_CS                                 ' Select chip
    HT_ShiftOut(cHT_ID_WR, 1 << 2)              ' Send ID: Write to RAM
    HT_ShiftOut(bHT_SureAddress, 1 << 6)        ' Send address
    HT_ShiftOut(bHT_SureOutValue, 1 << 3)       ' send 4 bits of data
    Set HT_CS                                   ' De-Select Chip
    Return
#endif
'-------------------------------------------------------------------------------------
' Set or clear a pixel
' 
' Input     : bHT_SetPixelX holds the X position of the pixel
'             bHT_SetPixelY holds the Y position of the Pixel
' Output    : None
' Notes     : Plot a point on the display, with the upper left hand corner being (0,0), and the lower right hand corner being (23, 15).
'             Y increases going "downward" in common with many displays
'
$define HT_Plot(pX, pY) bHT_SetPixelX = pX : WREG = pY : _HT_Plot
_HT_Plot Macro
    GoSub __HT_SetPixel
Endm
#ifdef _HT_Plot#Req
__HT_SetPixel:   
    bHT_SetPixelY = WREG                            ' Transfer WREG into it's working variable
    If WREG.7 = 1 Then Return                       ' \
    If bHT_SetPixelY > cHT_YMax Then Return         ' |
    If bHT_SetPixelX.7 = 1 Then Return              ' | Boundary checks
    If bHT_SetPixelX > cHT_XMax Then Return         ' /
'
' The 4 bits in a single memory word go down, with the LSB (last transmitted) bit being on top.  
' However, HT_ShiftOut() sends the LSB first, so we have to do a bit-reversal somewhere.
' This is done by shifting the single bit in the opposite direction from what you might expect.
'  
    bHT_BitVal = bHT_SetPixelY & 3                  ' \ Calculate which bit to act upon
    bHT_BitVal = 8 >> bHT_BitVal                    ' /
    bHT_Address = bHT_SetPixelY / 4                 ' \
    PRODL = bHT_SetPixelX * 4                       ' | Calculate which memory word this is in
    bHT_Address = bHT_Address + PRODL               ' /
    PRODL = bHT_ShadowRam[bHT_Address]               ' Speed things up by fetching from shadow RAM  
    If tPixelColor = tPixelSet Then                 ' Do we need to set a pixel?
        PRODH = PRODL | bHT_BitVal                  ' Yes. So set the pixel
    ElseIf tPixelColor = 2 Then                     ' Do we need to Xor the pixel?                      
        PRODH = PRODL ^ bHT_BitVal                  ' Yes. So Xor the pixel
    Else                                            ' Otherwise...
        WREG = ~bHT_BitVal                          ' \
        PRODH = PRODL & WREG                        ' / Clear the pixel
    EndIf
    bHT_ShadowRam[bHT_Address] = PRODH               ' Place the new value into the shadow RAM
    HT_SendData(bHT_Address, PRODH)                 ' Send the new memory value to the display
    Return 
#endif
'------------------------------------------------------------------------------------- 
' Read a pixel's state 
' 
' Input     : bHT_SetPixelX holds the X position of the pixel
'             bHT_SetPixelY holds the Y position of the Pixel
' Output    : WREG holds 1 or 0 depending on whether the pixel is set or clear
'             Returns with WREG holding 0 if out of bounds
' Notes     : Looks up the value of a pixel from the shadow RAM for speed 
'
$define HT_Pixel(pX, pY, pReturn) bHT_SetPixelX = pX : WREG = pY : _HT_Pixel : pReturn = WREG 
_HT_Pixel Macro
    GoSub __HT_GetPixel
Endm
#ifdef _HT_Pixel#Req
__HT_GetPixel:
    bHT_SetPixelY = WREG                               ' Transfer WREG into it's working variable
    WREG = 0                                           ' Clear return ready for a boundary overflow
    If bHT_SetPixelY.7 = 1 Then Return                 ' \
    If bHT_SetPixelY > cHT_YMax Then Return            ' | Boundary checks
    If bHT_SetPixelX.7 = 1 Then Return                 ' | Return holding 0 if out of bounds
    If bHT_SetPixelX > cHT_XMax Then Return            ' /
    
    bHT_BitVal = bHT_SetPixelY & 3                     ' \ Calculate which bit to act upon
    bHT_BitVal = 8 >> bHT_BitVal                       ' /
    bHT_Address = bHT_SetPixelY / 4                    ' \
    PRODL = bHT_SetPixelX * 4                          ' | Calculate which memory word this is in
    bHT_Address = bHT_Address + PRODL                  ' /  
    PRODL = bHT_ShadowRam[bHT_Address]                 ' Speed things up by fetching from shadow RAM
    WREG = PRODL & bHT_BitVal
    If WREG > 0 Then WREG = 1
    Return
#endif
'-------------------------------------------------------------------------------------
' Draw or Erase a line from bHT_Xpos_S, bHT_Ypos_S to bHT_Xpos_E, bHT_Ypos_E
'
' Input     : bHT_Xpos_S = Starting X position of the Line
'           : bHT_Ypos_S = Starting Y position of the line
'           : bHT_Xpos_E = Ending X position of the line
'           : bHT_Ypos_E = Endif Y position of the line
'           : tPixelColor = Colour of the pixel
' Output    : None
' Notes     :
'
$define HT_Line(pXS,pYS,pXE,pYE) bHT_Xpos_S = pXS : bHT_Ypos_S = pYS : bHT_Xpos_E = pXE : WREG = pYE : _HT_Line
_HT_Line Macro
    GoSub __HT_DrawLine
Endm
#ifdef _HT_Line#Req
Warnings = Off
Small_Micro_Model = On
__HT_DrawLine:
    bHT_Ypos_E = WREG                               ' Transfer WREG into it's working variable
    bHT_DeltaY = bHT_Ypos_E - bHT_Ypos_S            ' Find the initial Y direction
    bHT_DeltaX = bHT_Xpos_E - bHT_Xpos_S            ' Find the initial X direction

    bHT_StepY = 1                                   ' Default to a positive step for Ypos
    If bHT_DeltaY.7 = 1 Then                        ' Check if negative
           bHT_DeltaY = -bHT_DeltaY                 ' Negate bHT_DeltaY
           bHT_StepY = -1                           ' And indicate a negative step
    EndIf   
    bHT_StepX = 1                                   ' Default to a positive step for Xpos
    If bHT_DeltaX.7 = 1 Then                        ' Check if negative
           bHT_DeltaX = -bHT_DeltaX                 ' Negate bHT_DeltaX
           bHT_StepX = -1                           ' And indicate a negative step
    EndIf
    HT_Plot(bHT_Xpos_S,bHT_Ypos_S)                 ' The first pixel plot is a special case
    If bHT_DeltaX > bHT_DeltaY Then                    
           bHT_Deviation = bHT_DeltaY - bHT_DeltaX
        While bHT_Xpos_S <> bHT_Xpos_E
            If bHT_Deviation.7 = 0 Then             ' Check if negative
                bHT_Ypos_S = bHT_Ypos_S + bHT_StepY ' Move on the Y axis directed by bHT_StepY
                bHT_Deviation = bHT_Deviation - bHT_DeltaX
             EndIf
             bHT_Xpos_S = bHT_Xpos_S + bHT_StepX    ' Move on the X axis directed by bHT_StepX
             bHT_Deviation = bHT_Deviation + bHT_DeltaY
            HT_Plot(bHT_Xpos_S,bHT_Ypos_S)         ' Draw or Erase the pixel at this position
        Wend
        Return                                      ' Return prematurely (this saves an else command)
    EndIf
    bHT_Deviation = bHT_DeltaX - bHT_DeltaY        
            
    While bHT_Ypos_S <> bHT_Ypos_E
        If bHT_Deviation.7 = 0 Then                 ' Check if negative
            bHT_Xpos_S = bHT_Xpos_S + bHT_StepX     ' Move on the X axis directed by bHT_StepX
            bHT_Deviation = bHT_Deviation - bHT_DeltaY
        EndIf
        bHT_Ypos_S = bHT_Ypos_S + bHT_StepY         ' Move on the Y axis directed by bHT_StepY
        bHT_Deviation = bHT_Deviation + bHT_DeltaX
        HT_Plot(bHT_Xpos_S,bHT_Ypos_S)             ' Draw or Erase the pixel at this position
    Wend
    Return
Small_Micro_Model = Off 
#endif
'-------------------------------------------------------------------------------------
' Draw a circle
$define HT_Circle(pX, pY, pR) bHT_Xpos_S = pX : bHT_Ypos_S = pY : WREG = pR : _HT_Circle
_HT_Circle Macro
    GoSub __HT_Circle
Endm
#ifdef _HT_Circle#Req
__HT_Circle:
    bHT_Radius = WREG                              ' Transfer WREG into its working variable
    bHT_DD = bHT_Ypos_S - bHT_Xpos_S
    bHT_XX = 0
    bHT_YY = bHT_Radius
    bHT_TR = 3 - (2 * bHT_Radius)
    While bHT_XX <= bHT_YY
        bHT_QuadA = bHT_Xpos_S + bHT_XX
        bHT_QuadB = bHT_Ypos_S + bHT_YY
        bHT_QuadC = bHT_Xpos_S - bHT_XX
        bHT_QuadD = bHT_Ypos_S - bHT_YY
        GoSub HT_PlotIt
        bHT_QuadA = bHT_Ypos_S + bHT_YY - bHT_DD
        bHT_QuadB = bHT_Ypos_S + bHT_XX
        bHT_QuadC = bHT_Ypos_S - bHT_YY - bHT_DD
        bHT_QuadD = bHT_Ypos_S - bHT_XX
        GoSub HT_PlotIt
        If bHT_TR.7 = 1 Then
            bHT_TR = bHT_TR + 6
            bHT_TR = bHT_TR + (4 * bHT_XX)
        Else
            bHT_TR = bHT_TR + 10
            bHT_TR = bHT_TR + (4 * (bHT_XX - bHT_YY))
            Dec bHT_YY
        EndIf
        Inc bHT_XX
    Wend
    Return
'
' Plot 4 locations simultaneously
' This is part of the Circle subroutine
'
HT_PlotIt:
    HT_Plot(bHT_QuadA,bHT_QuadB)
    HT_Plot(bHT_QuadA,bHT_QuadD)
    HT_Plot(bHT_QuadC,bHT_QuadB)
    HT_Plot(bHT_QuadC,bHT_QuadD)
    Return
#endif
'-------------------------------------------------------------------------------------
'Copyacharacterglyphfromthefontdatatabletodisplaymemory,withitsupperleftatthegivencoordinate
'
' Input     : bHT_CharX holds the X position of the character
'             bHT_CharY holds the Y position of the character
'             WREG holds the character to display
' Output    :WREG still holds the character sent
'Notes     : UsesHT_Plottodraweachdot of the character
'
$define HT_DisplayChar(pX,pY,pCh) bHT_CharX = pX :bHT_CharY = pY :WREG = pCh : _HT_DisplayChar
_HT_DisplayChar Macro
    GoSub __HT_DisplayChar
Endm
#ifdef _HT_DisplayChar#Req
__HT_DisplayChar:
    PP3H = WREG                                 ' Transfer WREG into its working variable
    tHT_PixelColSave = tPixelColor               ' Save the value of tPixelColor
'
'Movecharacterdefinition,pixelbypixel,ontothedisplay
' Fontsaredefinedasonebytepercolumn
'
' Characters's Address =  PP3H * 6
' PP3H holds the byte to display                     
' Add the offset for the address of font: wCharOffset + font label's address
' This is the code that requires altering for different sized fonts
'    
    bHT_Col = 0
    Repeat                                      ' Form a loop for all the bytes in the character's font
        wHT_CharOffset = PP3H * 6               ' 6 bytes per font
        wHT_CharOffset = wHT_CharOffset + bHT_Col ' Add the offset to the character's value
        TABLAT = LRead8 Font[wHT_CharOffset]    ' Read the byte from the font table
        '
        ' Send the value in TABLAT to the Display
        '
        bHT_Row = 7
        Repeat
            tPixelColor = TABLAT.6
            HT_Plot(bHT_CharX + bHT_Col,bHT_CharY + bHT_Row)
            TABLAT = TABLAT << 1
            Dec bHT_Row
        Until bHT_Row = 0
        Inc bHT_Col                             ' Next byte in the font
    Until bHT_Col >= 5                          ' Have we done 5 writes yet ?
    WREG = PP3H                                 ' Restore the value of WREG
    tPixelColor = tHT_PixelColSave               ' Restore the value of tPixelColor
    Return
#endif
'-------------------------------------------------------------------------------------
' Display a string of characters on the LED matrix
'
' Input     : bHT_CharX holds the X position of the character
'             bHT_CharY holds the Y position of the character
'             SFR pair FSR1L\H hold the address of the string to display
' Output    : None
' Notes     : None
'
$define HT_DisplayString(pX,pY,pStr) bHT_CharX = pX :WREG = pY :wHT_FSR1 = VarPtr pStr : _HT_DisplayString
_HT_DisplayString Macro
    GoSub __HT_DisplayString
Endm
#ifdef _HT_DisplayString#Req
__HT_DisplayString:
    bHT_CharY = WREG                                ' Transfer WREG into its working variable
    bHT_StepX = 0                                   ' Reset a counter
    Repeat                                          ' Form a loop
        WREG = POSTINC1                             ' Transfer the element of the string into WREG and increment position
        Bz _HT_DisplayString_Exit                   ' Exit if a null found
        HT_DisplayChar(bHT_CharX, bHT_CharY, WREG)  ' Display the character held in WREG
        bHT_CharX = bHT_CharX + 6                   ' Move 6 pixels to the right (with of each character)
        Inc bHT_StepX                               ' Increment the counter
    Until bHT_StepX = 0                             ' Until 256 iterations have occured
_HT_DisplayString_Exit:
    Return
#endif
'-------------------------------------------------------------------------------------
' Clear the display
'
' Input     : None
' Output    : None
' Notes     : None
'
$define HT_Clear() _HT_Clear
_HT_Clear Macro
    GoSub __HT_Clear
Endm
#ifdef _HT_Clear#Req
__HT_Clear:
    Clear bHT_ShadowRam                          ' Clear the shadow RAM
    bHT_DeltaX = 0
    Repeat                                      ' Create a loop
        HT_SendData(bHT_DeltaX,0)               ' Clear all the displays!
        Inc bHT_DeltaX
    Until bHT_DeltaX >= 127                     ' Until all are done
    Return
#endif
'-------------------------------------------------------------------------------------
' Initialise the Sure 2416 display board
'
' Input     : None
' Output    : None
' Notes     : None
'
$define HT_Init() _HT_Init
_HT_Init Macro
    GoSub __HT_Init
Endm
#ifdef _HT_Init#Req
__HT_Init:
    High HT_CS                                  ' Disable the board's interface
    High HT_WR                                  ' \ Place the board in write mode to start with
    Output HT_RD                                ' /
    Output HT_Data
    HT_SendCommand(cHT_CMD_SYSDIS)              ' Disable system
    HT_SendCommand(cHT_CMD_COMS11)              ' 16*32, PMOS drivers
    HT_SendCommand(cHT_CMD_MSTMD)               ' Master Mode
    HT_SendCommand(cHT_CMD_SYSON)               ' System on
    HT_SendCommand(cHT_CMD_LEDON)               ' LEDs on 
    HT_Clear()                                  ' Clear the display
    tPixelColor = tPixelSet                         ' Default to the pixel being set
    Return
#endif
'-------------------------------------------------------------------------------------
' Load the font for character display
    Include "Font.Inc"

'-------------------------------------------------------------------------------------
HT_StartUp:
    HT_Init()                                   ' Initialise the display
  
    
    

