'
' Game of Life for use on the Sure 2416 LED dot matrix display
' Make sure you are using version 1.1 of the "2416.inc" file
'
    Include "2416.inc"                          ' Load the Sure 2416 display interface routines into the program
    
    Dim bDifferenceCount As Byte                  ' Indicates whether the display is static or not (0 if static)
    Dim bNeighbours As Byte                     ' Neighbour counter
    Dim wGenerationCount As Byte                ' Generation counter
    Dim Xpos As Byte                            ' X position of display and shadow display RAM
    Dim Ypos As Byte                            ' Y position of display and shadow display RAM
    Dim bNeighbourXpos As Byte                  ' X position of a neighbour cell
    Dim bNeighbourYpos As Byte                  ' Y position of a neighbour cell
    Dim wRandResult As Word                     ' Result of the pseudo random number generator
    Dim bNeighbourState As Byte                 ' The display pixel's condition (0 or 1)
    Dim tShadowPixelColor As Byte               ' The colour of the shadow display pixel (0, 1, or 2)
    
    Dim bShadowRam[96] As Byte                  ' Shadow display RAM
    
    Symbol cAmountOfGenerations = 220           ' The amount of generations to perform before re-seeding
    Symbol cXpos_Max = 23                       ' Maximum X position of display (0 to 23)
    Symbol cYpos_Max = 15                       ' Maximum Y position of display (0 to 15)
'-------------------------------------------------------------------------------------    
    GoTo Main                                   ' Jump over the subroutines
'-------------------------------------------------------------------------------------
' Set or clear a shadow display pixel
' 
' Input     : bHT_SetPixelX holds the X position of the pixel
'             bHT_SetPixelY holds the Y position of the Pixel
' Output    : None
' Notes     : Plot a point in the shadow display RAM
'             bHT_SetPixelX and bHT_SetPixelY are declared in the "2416.inc" file
'
$define Shadow_Plot(pX, pY) bHT_SetPixelX = pX : WREG = pY : GoSub _Shadow_Plot
_Shadow_Plot:   
    bHT_SetPixelY = WREG                                ' Transfer WREG into it's working variable
    If bHT_SetPixelY.7 = 1 Then Return                  ' \
    If bHT_SetPixelY > cHT_YMax Then Return             ' |
    If bHT_SetPixelX.7 = 1 Then Return                  ' | Boundary checks
    If bHT_SetPixelX > cHT_XMax Then Return             ' /
'
' The 4 bits in a single memory word go down, with the LSB (last transmitted) bit being on top.  
' However, HT_ShiftOut() sends the LSB first, so we have to do a bit-reversal somewhere.
' This is done by shifting the single bit in the opposite direction from what you might expect.
'  
    bHT_BitVal = bHT_SetPixelY & 3                      ' \ Calculate which bit to act upon
    bHT_BitVal = 8 >> bHT_BitVal                        ' /
    bHT_Address = bHT_SetPixelY / 4                     ' \
    PRODL = bHT_SetPixelX * 4                           ' | Calculate which memory word this is in
    bHT_Address = bHT_Address + PRODL                   ' /
    PRODL = bShadowRam[bHT_Address]                     ' fetch from shadow display RAM  
    If tShadowPixelColor = tPixelSet Then               ' Do we need to set a shadow pixel?
        PRODH = PRODL | bHT_BitVal                      ' Yes. So set the shadow pixel
    Else                                                ' Otherwise...
        WREG = ~bHT_BitVal                              ' \
        PRODH = PRODL & WREG                            ' / Clear the shadow pixel
    EndIf
    bShadowRam[bHT_Address] = PRODH                     ' Place the new value into the shadow RAM
    Return 
'------------------------------------------------------------------------------------- 
' Read a shadow display pixel's state 
' 
' Input     : bHT_SetPixelX holds the X position of the pixel
'             bHT_SetPixelY holds the Y position of the pixel
' Output    : WREG holds 1 or 0 depending on whether the shadow display's pixel is set or clear
'             Returns with WREG holding 0 if out of bounds
'             bHT_SetPixelX and bHT_SetPixelY are declared in the "2416.inc" file
'
$define Shadow_Pixel(pX, pY, pReturn) bHT_SetPixelX = pX : WREG = pY : GoSub _Shadow_Pixel : pReturn = WREG 
_Shadow_Pixel:
    bHT_SetPixelY = WREG                                ' Transfer WREG into it's working variable
    WREG = 0                                            ' Clear return ready for a boundary overflow
    If bHT_SetPixelY.7 = 1 Then Return                  ' \
    If bHT_SetPixelY > cHT_YMax Then Return             ' | Boundary checks
    If bHT_SetPixelX.7 = 1 Then Return                  ' | Return holding 0 if out of bounds
    If bHT_SetPixelX > cHT_XMax Then Return             ' /
    
    bHT_BitVal = bHT_SetPixelY & 3                      ' \ Calculate which bit to act upon
    bHT_BitVal = 8 >> bHT_BitVal                        ' /
    bHT_Address = bHT_SetPixelY / 4                     ' \
    PRODL = bHT_SetPixelX * 4                           ' | Calculate which memory word this is in
    bHT_Address = bHT_Address + PRODL                   ' /  
    PRODL = bShadowRam[bHT_Address]                     ' Fetch from shadow display RAM
    WREG = PRODL & bHT_BitVal
    If WREG > 0 Then WREG = 1
    Return
'----------------------------------------------------------------------------------
' Check for a stationary display
'
' Input     : None
' Output    : bDifferenceCount holds 0 if the display is static
' Notes     : None
'
CheckForStaticDisplay:
    bDifferenceCount = 0
    For Ypos = 0 To cYpos_Max                               ' Create a loop for the size of the Y axis
        For Xpos = 0 To cXpos_Max                           ' Create a loop for the size of the X axis
            HT_Pixel(Xpos, Ypos, tPixelColor)               ' Read from the display into tPixelColor
            Shadow_Pixel(Xpos, Ypos, bNeighbourState)       ' Read the shadow display into bNeighbourState
            If tPixelColor <> bNeighbourState Then          ' Are they both the same
                Inc bDifferenceCount                        ' No. So increment a difference counter
            EndIf
        Next                                                ' Close the x loop
    Next                                                    ' Close the Y loop
    Return   
'---------------------------------------------------------------------------------- 
' Generate a pseudo random pattern of pixels on the display
'
' Input     : None
' Output    : bShadowRam holds a copy of the actual display
' Notes     : None
'
Generate:
    wGenerationCount = 0
    For Ypos = 1 To cYpos_Max - 1                               ' Scan the Y position
        For Xpos = 1 To cXpos_Max - 1                           ' Scan the X position
            wRandResult = Random                                ' Generate a pseudo random value between 1 and 65535
            Select wRandResult
                Case 48000 To 56000                             ' Look for a particular value
                    tShadowPixelColor = 1                       ' \
                    Shadow_Plot(Xpos,Ypos)                      ' / Set the shadow RAM's pixel
                    HT_Plot(Xpos,Ypos)                          ' Set the actual display's pixel
            EndSelect
        Next
    Next
    Return 
'----------------------------------------------------------------------------------
' Transfer the shadow display RAM to the actual display
'
' Input     : bShadowRam
' Output    : None
' Notes     : None
' 
Transfer_Generation:
    For Ypos = 0 To cYpos_Max
        For Xpos = 0 To cXpos_Max
            Shadow_Pixel(Xpos, Ypos, tPixelColor)       ' Read from the shadow display into tPixelColor 
            HT_Plot(Xpos,Ypos)                          ' Plot the resulting pixel colour
        Next
    Next
    Return 
'---------------------------------------------------------------------------------- 
' Check a neighbour cell (pixel) for live status
'
' Input     : bHT_SetPixelX holds the X position of the pixel
'             bHT_SetPixelY holds the Y position of the Pixel
' Output    : bNeighbours is incremented if the cell is alive
' Notes       bHT_SetPixelX and bHT_SetPixelY are declared in the "2416.inc" file
'
$define CheckNeighbour(pXpos, pYpos) bHT_SetPixelX = pXpos : bHT_SetPixelY = pYpos : GoSub _CheckNeighbour
_CheckNeighbour:
    HT_Pixel(bHT_SetPixelX, bHT_SetPixelY, bNeighbourState) ' Read the cell
    If bNeighbourState = 1 Then Inc bNeighbours             ' Increment the Neighbours counter if alive
    Return
'---------------------------------------------------------------------------------- 
' Calculate a single generation of cells
'
' Input     : None
' Output    : bShadowRam holds a copy of the altered colonies
' Notes     : Any live cell with fewer than two live neighbours dies, as if caused by underpopulation.
'             Any live cell with more than three live neighbours dies, as if by overcrowding.
'             Any live cell with two or three live neighbours lives on to the next generation.
'             Any dead cell with exactly three live neighbours becomes a live cell.
'
Calculate_Generations:
    Inc wGenerationCount                                    ' Increment the generation counter
    For Ypos = 0 To cYpos_Max                               ' Scan the Y position
        For Xpos = 0 To cXpos_Max                           ' Scan the X position
            bNeighbours = 0                                 ' Clear the bNeighbours counter
            CheckNeighbour(Xpos - 1, Ypos - 1)              ' Read the cell above and to the left
            CheckNeighbour(Xpos - 1, Ypos)                  ' Read the cell left
            CheckNeighbour(Xpos - 1, Ypos + 1)              ' Read the cell above and to the right
            CheckNeighbour(Xpos, Ypos - 1)                  ' Read the cell to the left
            CheckNeighbour(Xpos, Ypos + 1)                  ' Read the cell below
            CheckNeighbour(Xpos + 1, Ypos - 1)              ' Read the cell below and to the left
            CheckNeighbour(Xpos + 1, Ypos)                  ' Read the cell below
            CheckNeighbour(Xpos + 1, Ypos + 1)              ' Read the cell below and to the right
            
            HT_Pixel(Xpos, Ypos, bNeighbourState)           ' Read the current location of the cell into bNeighbourState
            
            If bNeighbourState = 1 Then                         ' Is the current cell alive?
                If bNeighbours <> 3 And bNeighbours <> 2 Then   ' Yes. So are there two or three live neighbours
                    tShadowPixelColor = 0                       ' \ Yes. So the cell has to die
                    Shadow_Plot(Xpos,Ypos)                      ' / Place the outcome in the shadow display RAM
                EndIf
            Else                                                ' Otherwise. the cell must be dead
                If bNeighbours = 3 Then                         ' Does the cell have 3 live neighbours?
                    tShadowPixelColor = 1                       ' \ Yes. So resurrect the cell
                    Shadow_Plot(Xpos,Ypos)                      ' / Place the outcome in the shadow display RAM
                EndIf   
            EndIf
        Next
    Next
    Return   
'----------------------------------------------------------------------------------   
' The main program loop starts here  
Main:
    HT_Clear()                                          ' Clear the display
    Clear bShadowRam                                    ' Clear the shadow display RAM
    Seed $0445                                          ' Seed the pseudo random generator
    While 1 = 1                                         ' Create an infinite loop
        GoSub Generate                                  ' Generate a colony of cells
        Repeat                                          ' Create a loop for the amount of generations
            GoSub Transfer_Generation                   ' Transfer the shadow display to the actual display
            GoSub Calculate_Generations                 ' Calculate a single generation
            GoSub CheckForStaticDisplay                 ' Check for a static display
            If bDifferenceCount = 0 Then                ' Is the display static?
                DelayMS 500                             ' Yes. So delay for a while
                Break                                   ' Exit the loop
            EndIf
            DelayMS 200                                 ' Delay so we can see what's happening
        Until wGenerationCount >= cAmountOfGenerations  ' Until the amount of generations is done      
        DelayMS 300                                     ' A small delay
    Wend                                                ' Do it forever
    
    
    
        
